@@ -1,6 +1,6 @@
 /**************************************************************************
  * 
- * Copyright 2007 Tungsten Graphics, Inc., Cedar Park, Texas.
+ * Copyright 2007 VMware, Inc.
  * All Rights Reserved.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -18,7 +18,7 @@
  * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
  * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.
- * IN NO EVENT SHALL TUNGSTEN GRAPHICS AND/OR ITS SUPPLIERS BE LIABLE FOR
+ * IN NO EVENT SHALL VMWARE AND/OR ITS SUPPLIERS BE LIABLE FOR
  * ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
  * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
  * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
@@ -77,13 +77,6 @@ llvmpipe_bind_sampler_states(struct pipe_context *pipe,
    assert(shader < PIPE_SHADER_TYPES);
    assert(start + num <= Elements(llvmpipe->samplers[shader]));
 
-   /* Check for no-op */
-   if (start + num <= llvmpipe->num_samplers[shader] &&
-       !memcmp(llvmpipe->samplers[shader] + start, samplers,
-               num * sizeof(void *))) {
-      return;
-   }
-
    draw_flush(llvmpipe->draw);
 
    /* set the new samplers */
@@ -125,13 +118,6 @@ llvmpipe_set_sampler_views(struct pipe_context *pipe,
    assert(shader < PIPE_SHADER_TYPES);
    assert(start + num <= Elements(llvmpipe->sampler_views[shader]));
 
-   /* Check for no-op */
-   if (start + num <= llvmpipe->num_sampler_views[shader] &&
-       !memcmp(llvmpipe->sampler_views[shader] + start, views,
-               num * sizeof(struct pipe_sampler_view *))) {
-      return;
-   }
-
    draw_flush(llvmpipe->draw);
 
    /* set the new sampler views */
